from collections import namedtuple
import math
import pickle

import numpy as np
import numpy.ma as ma
import numpy.testing
import pandas as pd

import pytest
import retinapy.mea_noise as mea


FF_NOISE_PATTERN_PATH = './data/ff_noise.h5'
FF_SPIKE_RESPONSE_PATH = './data/ff_spike_response.pickle'
FF_SPIKE_RESPONSE_PATH_ZIP = './data/ff_spike_response.pickle.zip'
FF_RECORDED_NOISE_PATH = './data/ff_recorded_noise.pickle'
FF_RECORDED_NOISE_PATH_ZIP = './data/ff_recorded_noise.pickle.zip'


class ExperimentData:

    def __init__(self, rec_name, stimulus_pattern, response, recorded_stimulus,
            downsample_factor = 18):
        self.rec_name = rec_name
        self.stimulus_pattern = stimulus_pattern
        self.response = response
        self.recorded_stimulus = recorded_stimulus
        self.downsample_factor = downsample_factor
        self._rec_id = None
        self._stimulus = None
        self._stimulus_sample_rate = None
        self._sensor_sample_rate = None

    @property
    def stimulus(self) -> np.ndarray:
        if not self._stimulus:
            stimlus, freq = mea.decompress_stimulus(
                    self.stimulus_pattern, 
                    self.recorded_stimulus, 
                    self.rec_name,
                    self.downsample_factor)
            self._stimulus = stimlus
            self._stimulus_sample_rate = freq
        return self._stimulus

    @property
    def stimulus_sample_rate(self) -> float:
        if not self._stimulus_sample_rate:
            assert not self._stimulus
            self.stimulus
            assert self._stimulus_sample_rate
        return self._stimulus_sample_rate

    @property
    def rec_id(self) -> int:
        if not self._rec_id:
            self._rec_id = mea.recording_names(
                    self.response).index(self.rec_name)
            assert self._rec_id
        return self._rec_id

    @property
    def sensor_sample_rate(self) -> float:
        if not self._sensor_sample_rate:
            self._sensor_sample_rate = self.recorded_stimulus.xs(self.rec_name,
                    'Recording').iloc[0]['Sampling_Freq']
            assert self._sensor_sample_rate
        self._sensor_sample_rate
        return self._sensor_sample_rate

    @property
    def num_sensor_samples(self) -> int:
        row = self.recorded_stimulus.xs(self.rec_name, level='Recording')\
                .iloc[0]
        num_samples = row['End_Fr'] - row['Begin_Fr'] + 1
        return num_samples

    def spikes(self, cluster_id) -> np.ndarray:
        spikes = self.response.xs(self.rec_name, level='Recording')\
                .reset_index('Stimulus ID').loc[cluster_id]['Spikes']\
                .compressed()
        return spikes


def test_load_stimulus_pattern():
    noise = mea.load_stimulus_pattern(FF_NOISE_PATTERN_PATH)
    known_shape = (24000, 4)
    assert noise.shape == known_shape


def test_load_response():
    for path in (FF_SPIKE_RESPONSE_PATH, FF_SPIKE_RESPONSE_PATH_ZIP):
        response = mea.load_response(path)
        known_index_names = ['Cell index', 'Stimulus ID', 'Recording']
        assert response.index.names == known_index_names
        known_shape = (4417, 2)
        assert response.shape == known_shape


def test_load_recorded_stimulus():
    for path in (FF_RECORDED_NOISE_PATH, FF_RECORDED_NOISE_PATH_ZIP):
        res = mea.load_recorded_stimulus(path)
        known_index_names = ['Stimulus_index', 'Recording']
        assert res.index.names == known_index_names
        known_shape = (18, 8)
        assert res.shape == known_shape


@pytest.fixture
def stimulus_pattern():
    return mea.load_stimulus_pattern(FF_NOISE_PATTERN_PATH)


@pytest.fixture
def recorded_stimulus():
    return mea.load_recorded_stimulus(FF_RECORDED_NOISE_PATH)


@pytest.fixture
def response_data():
    return mea.load_response(FF_SPIKE_RESPONSE_PATH)


@pytest.fixture
def exp0(stimulus_pattern, recorded_stimulus, response_data):
	exp = ExperimentData('Chicken_04_08_21_Phase_01', stimulus_pattern,
			response_data, recorded_stimulus)
	return exp


@pytest.fixture
def exp12(stimulus_pattern, recorded_stimulus, response_data):
	exp = ExperimentData('Chicken_17_08_21_Phase_00', stimulus_pattern, 
            response_data, recorded_stimulus)
	return exp


@pytest.fixture
def decimated_stimulus_freq():
    stim, freq = mea.decompress_stimulus(
            mea.load_stimulus_pattern(FF_NOISE_PATTERN_PATH),
            mea.load_recorded_stimulus(FF_RECORDED_NOISE_PATH),
            'Chicken_17_08_21_Phase_00',
            18)
    return stim, freq


def test_recording_names(response_data):
    known_list = [
            'Chicken_04_08_21_Phase_01',
            'Chicken_04_08_21_Phase_02',
            'Chicken_05_08_21_Phase_00',
            'Chicken_05_08_21_Phase_01',
            'Chicken_06_08_21_2nd_Phase_00',
            'Chicken_06_08_21_Phase_00',
            'Chicken_11_08_21_Phase_00',
            'Chicken_12_08_21_Phase_00',
            'Chicken_12_08_21_Phase_02',
            'Chicken_13_08_21_Phase_00',
            'Chicken_13_08_21_Phase_01',
            'Chicken_14_08_21_Phase_00',
            'Chicken_17_08_21_Phase_00',
            'Chicken_19_08_21_Phase_00',
            'Chicken_19_08_21_Phase_01',
            'Chicken_20_08_21_Phase_00',
            'Chicken_21_08_21_Phase_00']  
    rec_list = mea.recording_names(response_data)
    assert rec_list == known_list


def test_cluster_ids(response_data):
    known_list = [12, 13, 14, 15, 17, 25, 28, 29, 34, 44, 45, 50, 60, 61, 80,
                  82, 99, 114, 119, 149, 217, 224, 287, 317, 421, 553, 591]
    recording_name = 'Chicken_21_08_21_Phase_00'
    cluster_ids = mea.cluster_ids(response_data, recording_name)
    assert cluster_ids == known_list


def test_decompress_stimulus(stimulus_pattern, recorded_stimulus):
    # Test 1
    # No downsampling.
    decomp, new_freq = mea.decompress_stimulus(
            stimulus_pattern, recorded_stimulus,
            'Chicken_17_08_21_Phase_00',
            downsample_factor=1)
    known_length = 16071532
    assert decomp.shape == (known_length, mea.NUM_STIMULUS_LEDS)
    assert new_freq == pytest.approx(mea.ELECTRODE_FREQ)

    # Test 2
    # Downsample at x18.
    downsample_factor = 18
    orig_len = 16071532
    expected_decimated_len = math.ceil(orig_len / downsample_factor)
    orig_freq = mea.ELECTRODE_FREQ
    expected_downsampled_freq = orig_freq / downsample_factor
    decomp, new_freq = mea.decompress_stimulus(
            stimulus_pattern,
            recorded_stimulus,
            'Chicken_17_08_21_Phase_00',
            downsample_factor)
    assert decomp.shape == (expected_decimated_len, mea.NUM_STIMULUS_LEDS)
    # Note: the sampling frequency in the Pandas dataframe isn't as accurate
    # as the ELECTRODE_FREQ value.
    assert new_freq == pytest.approx(expected_downsampled_freq)


def test_downsample_stimulus():
    # Setup
    orig_signal = np.array([0, 0, 0, 0, 0, 0, 0, 0,
                            1, 1, 1, 1, 1, 1, 1, 1,
                            0, 0, 0, 0, 0, 0, 0, 0,
                            1, 1, 1, 1, 1, 1, 1, 1,
                            0, 0, 0, 0, 0, 0, 0, 0,
                            1, 1, 1, 1, 1, 1, 1, 1])
    # TODO: is this satisfactory?
    expected_decimate_by_2 = np.array(
            [0.012, -0.019,  0.029, -0.060,  
             0.739,  1.085,  0.937,  1.081,  
             0.249, -0.076,  0.058, -0.077,
             0.750,  1.077,  0.942,  1.077,  
             0.250, -0.077,  0.058, -0.076,  
             0.749,  1.081,  0.937,  1.085])
    expected_decimate_by_4 = np.array(
            [0.030, -0.065,  0.595,  1.148,  
             0.364, -0.120,  0.620,  1.136,  
             0.369, -0.120,  0.614,  1.148])

    # Test
    decimated_by_2 = mea.downsample_stimulus(orig_signal, 2)
    numpy.testing.assert_allclose(decimated_by_2, 
            expected_decimate_by_2, atol=0.002)
    decimated_by_4 = mea.downsample_stimulus(orig_signal, 4)
    numpy.testing.assert_allclose(decimated_by_4, expected_decimate_by_4,
            atol=0.002)


def test_decompress_stimulus():
    # Setup
    stimulus_pattern = np.array([
        [1, 1, 1, 1, 0, 0, 0, 0],
        [1, 1, 0, 0, 1, 1, 0, 0],
        [1, 0, 1, 0, 1, 0, 1, 0]]).T
    trigger_events = np.array([0, 4, 5, 7, 12])
    # Out array should be overridden, so check this by filling with some value.
    out_arr = np.full(shape=(20, stimulus_pattern.shape[1]), fill_value=7)
    expected_output = np.array([
        # idx: 0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18 19
        # val: 0  0  0  0  1  2  2  3  3  3  3  3  4  4  4  4  4  4  4  4
              [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
              [1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1],
              [1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]]).T

    # Test 1
    mea._decompress_stimulus(stimulus_pattern, trigger_events, out_arr)
    np.testing.assert_array_equal(out_arr, expected_output)

    # Test 2
    # Non-zero starting trigger is invalid.
    trigger_events_invalid = np.array([2, 4, 5, 7, 12])
    with pytest.raises(ValueError):
        mea._decompress_stimulus(stimulus_pattern, trigger_events_invalid,
                out_arr)


def test_decompress_spikes1():
    # Setup
    downsample_by = 9
    num_sensor_samples = 123
    spike_times1 = np.array([0, 1, 8, 9, 10, 27, 30, 40, 50, 70, 80, 90, 100, 110])
    spike_times2 = np.array([8, 9, 30, 40, 50, 70, 80, 90, 100, 110])

    # input index:          [0, 9, 18, 27, 36, 45, 54, 63, 72, 81, 90, 99, 108, 117]
    # output index          [0, 1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,  12,  13]
    spike_counts1 = np.array([3, 2,  0,  2,  1,  1,  0,  1,  1,  0,  1,  1,  1,  0])
    spike_counts2 = np.array([1, 1,  0,  1,  1,  1,  0,  1,  1,  0,  1,  1,  1,  0])

    expected_output_len = math.ceil(num_sensor_samples / downsample_by)
    expected_ones_mask = np.zeros(expected_output_len, dtype=bool)
    expected_ones_mask[spike_times1 // downsample_by] = True
    expected_zeros_mask = ~expected_ones_mask

    # Test 1
    # There should be an error thrown, as two samples land in the same bucket.
    with pytest.raises(ValueError):
        mea.decompress_spikes(spike_times1, num_sensor_samples, downsample_by)

    # Test 2
    spikes = mea.decompress_spikes(spike_times2, num_sensor_samples, 
            downsample_by)
    assert spikes.shape == (expected_output_len, )
    assert np.all(spikes[expected_ones_mask] == True)
    assert np.all(spikes[expected_zeros_mask] == False)


def test_decompress_spikes2(exp12):
    """
    Test decompress_spikes on actual recordings.

    Not much is actually checked though.
    """
    # Setup
    cluster_id = 36
    spikes = exp12.spikes(cluster_id)
    assert spikes.shape == (2361,)

    # Test 
    spikes_decomp1 = mea.decompress_spikes(spikes, exp12.num_sensor_samples)
    spikes_decomp18 = mea.decompress_spikes(spikes, exp12.num_sensor_samples,
            downsample_factor=18)


def test_spike_snippets():
    """
    Test stimulus_slice function.

    The focus is on testing the shape, position and padding of the slice.

    Note that the slicing function doesn't do any filtering, so we can 
    use numpy.assert_equals, as the array values will not be modified.
    """
    # Setup
    # -----
    stim_frame_of_spike = [4, 1, 0, 6, 7]
    # The numbers in comments refer to the 5 tests below.
    stimulus = np.array(
            [[1, 1, 1, 1], #     |  2
             [1, 1, 1, 1], #     1  |
             [0, 1, 1, 1], #  -  |  - 
             [0, 0, 1, 1], #  |  - 
             [0, 0, 0, 1], #  0       -
             [0, 0, 0, 0], #  |       |  -
             [1, 0, 0, 0], #  -       3  |
             [1, 1, 0, 0]] #          |  4
        )
    total_len = 5
    pad = 2
    stimulus_sample_rate = 40
    # A bit of reverse engineering to get the sensor frame of the spike.
    def stimulus_frame_to_spikes(stimulus_frame):
        frame_width_in_sensor_samples = \
            mea.ELECTRODE_FREQ / stimulus_sample_rate
        first_spike = stimulus_frame * frame_width_in_sensor_samples 
        spikes_in_frame = np.arange(first_spike, 
                first_spike + frame_width_in_sensor_samples)
        return spikes_in_frame

    # Test 0
    # ------
    # Case where no padding is needed.
    snippets = mea.spike_snippets(
            stimulus,
            stimulus_frame_to_spikes(stim_frame_of_spike[0]),
            stimulus_sample_rate,
            mea.ELECTRODE_FREQ,
            total_len,
            pad)
    for s in snippets:
        expected_slice = np.array([
            [0, 1, 1, 1],
            [0, 0, 1, 1],
            [0, 0, 0, 1],
            [0, 0, 0, 0],
            [1, 0, 0, 0]])
        assert s.shape == expected_slice.shape
        numpy.testing.assert_allclose(s, expected_slice)

    # TODO: shouldn't they be filtered?

    # Test 1
    # ------
    # Sample is near the beginning and needs padding.
    snippets = mea.spike_snippets(
            stimulus,
            stimulus_frame_to_spikes(stim_frame_of_spike[1]),
            stimulus_sample_rate,
            mea.ELECTRODE_FREQ,
            total_len,
            pad)
    for s in snippets:
        expected_slice = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 1, 1, 1],
            [0, 0, 1, 1]])
        assert s.shape == expected_slice.shape
        numpy.testing.assert_allclose(s, expected_slice)

    # Test 2
    # ------
    # Sample is _at_ the beginning and needs padding.
    snippets = mea.spike_snippets(
            stimulus,
            stimulus_frame_to_spikes(stim_frame_of_spike[2]),
            stimulus_sample_rate,
            mea.ELECTRODE_FREQ,
            total_len,
            pad)
    for s in snippets:
        expected_slice = np.array([
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 1, 1, 1]])
        assert s.shape == expected_slice.shape
        numpy.testing.assert_allclose(s, expected_slice)

    # Test 3
    # ------
    # Sample is near the end and needs padding.
    snippets = mea.spike_snippets(
            stimulus,
            stimulus_frame_to_spikes(stim_frame_of_spike[3]),
            stimulus_sample_rate,
            mea.ELECTRODE_FREQ,
            total_len,
            pad)
    for s in snippets:
        expected_slice = np.array([
            [0, 0, 0, 1],
            [0, 0, 0, 0],
            [1, 0, 0, 0],
            [1, 1, 0, 0],
            [0, 0, 0, 0]])
        assert s.shape == expected_slice.shape
        numpy.testing.assert_allclose(s, expected_slice)

    # Test 4
    # ------
    # Sample is _at_ the end and needs padding.
    snippets = mea.spike_snippets(
            stimulus,
            stimulus_frame_to_spikes(stim_frame_of_spike[4]),
            stimulus_sample_rate,
            mea.ELECTRODE_FREQ,
            total_len,
            pad)
    for s in snippets:
        expected_slice = np.array([
            [0, 0, 0, 0],
            [1, 0, 0, 0],
            [1, 1, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0]])
        assert s.shape == expected_slice.shape
        numpy.testing.assert_equal(s, expected_slice)


def test_save_recording_names(tmp_path, response_data):
    rec_names = mea.recording_names(response_data)
    path = mea._save_recording_names(rec_names, tmp_path)
    expected_path = tmp_path / mea.REC_NAMES_FILENAME
    assert path == expected_path
    assert expected_path.is_file()
    with open(expected_path, 'rb') as f:
        contents = pickle.load(f)
    assert contents == rec_names


def test_save_cluster_ids(tmp_path, response_data):
    rec_id = 3
    cluster_ids = mea.cluster_ids(response_data, 
            mea.recording_names(response_data)[rec_id])
    path = mea._save_cluster_ids(cluster_ids, rec_id, tmp_path)
    expected_path = tmp_path / str(rec_id) / mea.CLUSTER_IDS_FILENAME
    assert path == expected_path
    assert expected_path.is_file()
    with open(expected_path, 'rb') as f:
        contents = pickle.load(f)
    assert contents == cluster_ids


def test_labeled_spike_snippets():
    """
    Create a fake response `DataFrame` and check that the spike snippets are
    calculated correctly.
    """
    # Setup
    snippet_len = 7
    snippet_pad = 2
    stim_sample_freq = 40
    # Fake stimulus. Note that we are did our own nieve upsampling here. 
    # This is done to make the comparison easier.
    stimulus = np.array([
            [0,0,0,1], # 0
            [0,0,1,0], # 1
            [0,0,1,1], # 2
            [0,1,0,0], # 3
            [0,1,0,1], # 4
            [0,1,1,0], # 5
            [0,1,1,1], # 6
            [1,0,0,0], # 7
            [1,0,0,1], # 8
            [1,0,1,0], # 9
            [1,0,1,1], # 10
            [1,1,0,0], # 11
            ])
    # Fake response
    rec_name1 = 'Chicken_04_08_21_Phase_01'
    rec_name2 = 'Chicken_04_08_21_Phase_02'
    index = pd.MultiIndex.from_tuples(
            ((25, 1, 'Chicken_04_08_21_Phase_01'),
            (40, 1, 'Chicken_04_08_21_Phase_01'),
            (17, 1, 'Chicken_04_08_21_Phase_02'),
            (40, 1, 'Chicken_04_08_21_Phase_02')), 
            names=['Cell index', 'Stimulus ID', 'Recording'])
    # Spike data has format: [(Kernel, Spikes),...]
    # We only care about the spikes, so we can leave the kernels as None.
    spike_data = [(None, [820, 3648]), (None, [3044, 4067]), (None, [5239,]), 
            (None, [4430,])]
    # Convert the data to use masked arrays.
    spike_data_m = [(kernel, ma.array(spikes, mask=np.zeros(len(spikes)))) 
        for kernel, spikes in spike_data]

    response = pd.DataFrame(spike_data_m, index=index, 
            columns=['Kernel', 'Spikes'])
    # The following is the predicted snippets.
    expected_spike_snippets1 = np.array([
            [
                [0,0,0,0],  # pad
                [0,0,0,0],  # pad
                [0,0,0,0],  # pad
                [0,0,0,1],  # 0
                [0,0,1,0],  # 1 <-- spike
                [0,0,1,1],  # 2 
                [0,1,0,0],  # 3 
            ],         
            [          
                [0,1,0,1], # 4
                [0,1,1,0], # 5
                [0,1,1,1], # 6
                [1,0,0,0], # 7
                [1,0,0,1], # 8 <- spike
                [1,0,1,0], # 9
                [1,0,1,1], # 10
            ],         
            [          
                [0,0,1,1], # 2 
                [0,1,0,0], # 3
                [0,1,0,1], # 4
                [0,1,1,0], # 5
                [0,1,1,1], # 6 <- spike
                [1,0,0,0], # 7
                [1,0,0,1], # 8
            ],         
            [          
                [0,1,1,0], # 5
                [0,1,1,1], # 6
                [1,0,0,0], # 7
                [1,0,0,1], # 8
                [1,0,1,0], # 9 <- spike
                [1,0,1,1], # 10
                [1,1,0,0], # 11
            ]
        ])         
    expected_spike_snippets2 = np.array([
            [          
                [1,0,0,0], # 7
                [1,0,0,1], # 8
                [1,0,1,0], # 9
                [1,0,1,1], # 10
                [1,1,0,0], # 11 <- spike
                [0,0,0,0], # pad
                [0,0,0,0], # pad
            ],                 
            [                  
                [0,1,1,0], # 5
                [0,1,1,1], # 6
                [1,0,0,0], # 7
                [1,0,0,1], # 8
                [1,0,1,0], # 9 <- spike
                [1,0,1,1], # 10
                [1,1,0,0], # 11
            ]
        ])
    expected_cluster_ids1 = np.array([25, 25, 40, 40])
    expected_cluster_ids2 = np.array([17, 40])

    # The spike times were calculated according to the following function.
    # Then, the slices of the stimulus were manually copied, and zoomed by 2,
    # as the stimulus was sampled at 40 Hz (twice the stimulus frequency).
    # Leaving the function here in case the test arrays need to be regenerated.
    def create_ans():
        def spike_to_frame(sp):
            return sp * (stim_sample_freq / mea.ELECTRODE_FREQ)
        res = []
        for d in spike_data:
            d_spikes = d[1]
            for sp in d_spikes:
                res.append(spike_to_frame(sp))
        print(res)

    # Test 1 (rec_name1)
    spike_snippets, cluster_ids = mea.labeled_spike_snippets(stimulus, 
            response, rec_name1, stim_sample_freq, mea.ELECTRODE_FREQ, 
            snippet_len, snippet_pad)
    for idx, (spwin, cluster_ids) in enumerate(
            zip(spike_snippets, cluster_ids)):
        np.testing.assert_equal(spwin, expected_spike_snippets1[idx])
        np.testing.assert_equal(cluster_ids, expected_cluster_ids1[idx])

    # Test 2 (rec_name2)
    spike_snippets, cluster_ids = mea.labeled_spike_snippets(stimulus, 
            response, rec_name2, stim_sample_freq, mea.ELECTRODE_FREQ, 
            snippet_len, snippet_pad)
    for idx, (spwin, cluster_ids) in enumerate(
            zip(spike_snippets, cluster_ids)):
        np.testing.assert_equal(spwin, expected_spike_snippets2[idx])
        np.testing.assert_equal(cluster_ids, expected_cluster_ids2[idx])


def test_write_rec_snippets(tmp_path, exp12):
    # Setup
    snippet_len = 7
    snippet_pad = 1
    empty_snippets = 0
    snippets_per_file = 100

    # Test
    # TODO: add some more checks.
    # Currently, the test only checks that the method runs to completion.
    mea._write_rec_snippets(exp12.stimulus, exp12.response, exp12.rec_name, 
            exp12.rec_id, tmp_path, snippet_len, snippet_pad, 
			exp12.stimulus_sample_rate, empty_snippets, snippets_per_file)




from collections import namedtuple
from collections import defaultdict
import itertools
import logging
import math
import pathlib
import pickle
import tempfile
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import scipy
import scipy.signal

import h5py


ELECTRODE_FREQ = 17852.767845719834  # Hz
NUM_STIMULUS_LEDS = 4
REC_NAMES_FILENAME = 'recording_names.pickle'
CLUSTER_IDS_FILENAME = 'cluster_ids.pickle'
IDS_FILE_PATTERN = '{part}_ids.npy'
SNIPPET_FILE_PATTERN = '{part}_snippets.npy'
RNG_SEED = 123

# Named tuple for stimuli.
Stimulus = namedtuple('Stimulus', ['name', 'display_hex', 'import_name'])
stimuli = [
    Stimulus('red',   '#ff0a0a', '/Red_Noise'),
    Stimulus('green', '#0aff0a', '/Green_Noise'),
    Stimulus('blue',  '#0a0aff', '/Blue_Noise'),
    Stimulus('uv',    '#303030', '/UV_Noise')]
stimulus_names = tuple(s.name for s in stimuli)


def load_stimulus_pattern(file_path: str) -> np.ndarray:
    """
    Loads the stimulus data from the HDF5 file as a Pandas DataFrame.

    Dataframe structure
    -------------------
        - integer index, representing stimulus frames.
        - columns are ['red', 'green', 'blue', 'uv']
        - values are 0 or 1 representing whether the corresponding LED is ON
          or OFF at the given stimulus frame.
    """
    with h5py.File(file_path, 'r') as f:
        # The data has shape: [4, 24000, 10374]. This corresponds to 4 lights,
        # on-off pattern for 20min at 20 Hz (24000 periods), and 10374 boxes
        # arranged over 2D screen space. In the full-field experiment, only a
        # single box was used, and hence the [:,0] access pattern.
        colour_noise = np.array([
            f[stimuli[0].import_name][:, 0],
            f[stimuli[1].import_name][:, 0],
            f[stimuli[2].import_name][:, 0],
            f[stimuli[3].import_name][:, 0]]).transpose()
    stimulus_switch_freq = 20
    stimulus_loop_mins = 20
    stimulus_loop_secs = stimulus_loop_mins * 60
    assert colour_noise.shape[0] == stimulus_loop_secs * stimulus_switch_freq
    # Optional, if bottleneck encountered
    # ===================================
    # We are doing a lot of slicing in the time-step dimension, so keep it
    # as the row dimension. Also insure the array is contiguous for further
    # speed-ups.
    # res = np.ascontiguousarray(colour_noise)
    return colour_noise


def load_response(file_path: str, keep_kernels=False) -> pd.DataFrame:
    """Loads the spike data from a Pickle file as a Pandas DataFrame.

    The input path should point to standard pickle file (zipped or not).

    Dataframe structure
    -------------------
    The dataframe uses a multiindex: ['Cell index', 'Stimulus ID', 'Recording'].
        - The cell index refers to the id assigned by the spike sorter. 
        - TODO: The stimulus ID is what?
        - The 'Recording' index is a human readable string to identify a 
            recording session.

    The Dataframe has 2 columns: ['Kernel', 'Spikes'].
        - The each row contains a (2000,4) shaped Numpy array holding the
          precomputed response kernel for the matching cell-stimulus-recording.
          The kernel represents 2000 miliseconds, with the spike located at
          1000 ms. 
        - The 'Spikes' column contains a (17993,) shaped *masked* Numpy array,
          holding a variable number of integers. Each integer reprents the
          time (in sensor readings) at which the spike occurred.
    """
    res = pd.read_pickle(file_path)
    if not keep_kernels:
        res.drop('Kernel', axis=1)
    return res


def load_recorded_stimulus(file_path: str) -> pd.DataFrame:
    res = pd.read_pickle(file_path, compression='infer')
    return res


def decompress_stimulus(stimulus_pattern: np.ndarray, 
        stimulus_df: pd.DataFrame, rec_name: str, 
        downsample_factor: int = 1) -> Tuple[np.ndarray, float]:
    row = stimulus_df.xs(rec_name, level='Recording')
    if len(row) != 1:
        raise ValueError('Expected exactly one row for recording ({rec_name}).'
                         f' Got ({len(row)}).')
    row = row.iloc[0]
    # TODO: is 'End_Fr' inclusive? If so, add 1 below. Assuming yes for now.
    num_samples = row['End_Fr'] - row['Begin_Fr'] + 1
    if num_samples != int(num_samples):
        raise ValueError('Expected integer number of samples. Got '
                         f'({num_samples}).')
    trigger_events = row['Trigger_Fr_relative']
    # Two very similar code paths:
    res = np.empty((num_samples, stimulus_pattern.shape[1]))
    _decompress_stimulus(stimulus_pattern, trigger_events, res)
    res = downsample_stimulus(res, downsample_factor)

    # Optional, if bottleneck encountered
    # ===================================
    # We are doing a lot of slicing in the time-step dimension, so keep it
    # as the row dimension. Also insure the array is contiguous for further
    # speed-ups.
    # res = np.ascontiguousarray(res)

    # Note: the sampling frequency in the Pandas dataframe isn't as accurate
    # as the ELECTRODE_FREQ value.
    new_freq = row['Sampling_Freq'] / downsample_factor
    return res, new_freq


def _decompress_stimulus(
        stimulus_pattern: np.ndarray,
        trigger_events: np.ndarray,
        out_array: np.ndarray):
    """TODO: take in the stimulus trigger points."""
    logging.info('Starting: decompressing stimulus. Resulting shape '
                 f'({out_array.shape}).')
    if trigger_events[0] != 0:
        raise ValueError('The trigger events are expected to start at zero, '
                f'but the first trigger was at ({trigger_events[0]}).')
    if len(trigger_events) > len(stimulus_pattern):
        raise ValueError('Recorded stimulus is longer than the stimulus '
                f'pattern. ({len(trigger_events)} > {len(stimulus_pattern)})')
    # If this becomes a bottleneck, there are some tricks to reach for:
    # https://stackoverflow.com/questions/60049171/fill-values-in-numpy-array-that-are-between-a-certain-value
    slices = np.stack((trigger_events[:-1], trigger_events[1:]), axis=1)
    for idx, s in enumerate(slices):
        out_array[np.arange(*s)] = stimulus_pattern[idx]
    last_trigger = trigger_events[-1]
    out_array[last_trigger:] = stimulus_pattern[len(slices)]
    logging.info(f'Finished: decompressing stimulus. '
                 f'The last trigger was at ({trigger_events[-1]}) making its '
                 f'duration ({out_array.shape[0]} - {trigger_events[-1]} = '
                 f'{out_array.shape[0] - trigger_events[-1]}) samples.')
    return out_array


def decompress_spikes(
        spikes: Union[np.ndarray, np.ma.MaskedArray], 
        num_sensor_samples: int, 
        downsample_factor: int = 1) -> np.ndarray:
    """
    Fills an True/False array depending on whether a spike occurred.

    The output has one element for each downsampled timestep.
    """
    if np.ma.isMaskedArray(spikes):
        spikes = spikes.compressed()
    downsampled_spikes = np.floor_divide(spikes, downsample_factor)
    if np.min(np.diff(downsampled_spikes)) < 1:
        raise ValueError('Multiple spikes occured in the same bucket; spikes '
                'are happening faster than the sampling rate.')
    res = np.zeros(shape=[math.ceil(num_sensor_samples/downsample_factor),],
            dtype=bool)
    res[downsampled_spikes] = True
    return res


def factors_sorted_by_count(n, limit) -> List[Tuple[int,...]]:
    """
    Calulates factor decomposition with sort and limit.

    This method is used to choose downsampling factors when a single factor
    is too large. The decompositions are sorted by the number of factors in a
    decomposition.  
    """
    def _factors(n):
        res = [(n,)]
        f1 = n // 2
        while(f1 > 1):
            f2, mod = divmod(n,f1)
            if not mod:
                res.append((f1, f2)) 
                sub_factors = [a+b for (a,b) in
                        itertools.product(_factors(f1), _factors(f2))]
                res.extend(sub_factors)
            f1 -= 1
        return res
    factors = list(set(_factors(n)))
    factors_under = [f for f in factors if max(f) <= limit]
    sorted_by_count = sorted(factors_under, key = lambda x : len(x))
    return sorted_by_count


def downsample_stimulus(stimulus: np.ndarray, factor: int) -> np.ndarray:
    """
    Filter (low-pass) a stimulus and then decimate by a factor.

    This is needed to prevent aliasing.

    Resources on filtering
    ----------------------
    https://dsp.stackexchange.com/questions/45446/pythons-tt-resample-vs-tt-resample-poly-vs-tt-decimate
    https://dsp.stackexchange.com/questions/83696/downsample-a-signal-by-a-non-integer-factor
    https://dsp.stackexchange.com/questions/83889/decimate-a-signal-whose-values-are-calculated-not-stored?noredirect=1#comment176944_83889
    """
    if factor == 1:
        return stimulus
    time_axis = 0
    logging.info('Starting: downsampling')
    # SciPy recommends to never exceed 13 on a single decimation call.
    # See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.decimate.html
    MAX_SINGLE_DECIMATION = 13
    sub_factors = factors_sorted_by_count(factor, 
            limit=MAX_SINGLE_DECIMATION)[0]
    for sf in sub_factors:
        logging.info(f'Starting: decimating by {sf}')
        stimulus = scipy.signal.decimate(stimulus, sf, ftype='fir',
                axis=time_axis)
        logging.info(f'Finished: decimating by {sf}')
    logging.info('Finished: downsampling')
    return stimulus


def recording_names(response: pd.DataFrame) -> list:
    """Return the list of recording names."""
    rec_list = response.index.get_level_values('Recording').unique().tolist()
    return rec_list


def _list_to_index_map(l) -> Dict[str, int]:
    """
    Returns a map from list elements to their index in the list.
    """
    return {x: i for i, x in enumerate(l)}


def cluster_ids(response: pd.DataFrame, rec_name: str) -> list:
    """Returns the list of cluster ids for a given recording.

    Stimulus ID = 7 is currently ignored (actually, anything other than 1).
    """
    # Note: I'm not sure if it's better to request by recording name or
    # the recording id.
    stimulus_id = 1  # not 7.
    cluster_ids = response.xs((stimulus_id, rec_name), 
            level=('Stimulus ID', 'Recording'), drop_level=True) \
                    .index.get_level_values('Cell index').unique().tolist()
    return cluster_ids


def spike_window(spikes: Union[int, np.ndarray], total_len: int,
                 post_spike_len: int, stimulus_sample_rate: float,
                 sensor_sample_rate: float):
    """
    Calculate the window endpoints around a spike, in samples of the stimulus.
    """
    if sensor_sample_rate < stimulus_sample_rate:
        logging.warn(f'The sensor sample rate ({sensor_sample_rate}) is lower '
                     f'than the stimulus sample rate ({stimulus_sample_rate}).')
    if total_len < post_spike_len + 1:
        raise ValueError(f'Total length must be at least 1 greater than the '
                         f'post-spike lengths + 1. Got total_len ({total_len}) '
                         f'and post_spike_len ({post_spike_len}).')
    # 1. Select the sample to represent the spike.
    stimulus_frame_of_spike =  \
        np.floor(spikes * (stimulus_sample_rate / sensor_sample_rate)) \
        .astype('int')
    # 2. Calculate the snippet start and end.
    # The -1 appears as we include the spike sample in the snippet.
    win_start = (stimulus_frame_of_spike + post_spike_len) - (total_len - 1)
    win_end = stimulus_frame_of_spike + post_spike_len + 1
    return win_start, win_end


def spike_snippets(stimulus: np.ndarray, spikes: np.ndarray, 
                   stimulus_sample_rate: float,
                   sensor_sample_rate: float, total_len: int, 
                   post_spike_len: int) -> np.ndarray:
    """
    Return a subset of the stimulus around the spike points.

    Args:
        stimulus: The stimulus.
        spikes: The sensor samples in which the spikes were detected.
        total_len: The length of the snippet in stimulus frames. 
        post_spike_len: The number of frames to pad after the spike.
        stimulus_sample_rate: The sampling rate of the stimulus.
        sensor_sample_rate: The sampling rate of the sensor.
    Returns:
        A Numpy array of shape (spikes.shape[0], total_len, NUM_STIMULUS_LEDS).

    Note 1: The total length describes the snippet length inclusive of the post
        spike padding.
    Note 2: If the spike falls exactly on the boundary of stimulus samples,
            then the earlier stimulus sample will be used.

    Example
    =======

        frame #:  |   0   |   1   |   2   |   3   |   4   |   5   |   6   |   7   |
        ===========================================================================
        stim:     |   0   |   1   |   1   |   0   |   0   |   1   |   0   |   0   |
                  |   1   |   1   |   0   |   0   |   1   |   0   |   0   |   0   |
                  |   1   |   1   |   1   |   0   |   0   |   0   |   0   |   1   |
                  |   0   |   0   |   0   |   0   |   1   |   0   |   1   |   1   |
        ===========================================================================
        resp:     |----------------------------------*----------------------------|

    The slice with parameters:
        - total length = 6
        - post spike length = 1

    Would be:

                  |   1   |   1   |   0   |   0   |   1   |              
                  |   1   |   0   |   0   |   1   |   0   |              
                  |   1   |   1   |   0   |   0   |   0   |              
                  |   0   |   0   |   0   |   1   |   0   |              


    Frame calculation
     ----------------
    (I'm not 100% confident that the below sample choice is correct.)

    In the following scenario:

        stimulus frame:  |  0  |  1  |  2  |  3  |
        spike:           |-----*-----------------|

    The stimulus frame in which the spike falls is:

        stimulus frame of spike = floor(spike_frame * ELECTRODE_FREQ / 
                                                       stimulus_sampling_rate)
                                = 1  (in the above example)

    In the following scenario all of the following spikes fall within the same 
    stimulus frame:

        stimulus frame: |   0   |   1   |   2   |
        spike time:     |-------********--------|
    """
    # 1. Get the spike windows.
    win_start, _ = spike_window(spikes, total_len, post_spike_len,
                                stimulus_sample_rate, sensor_sample_rate)
    # 2. Pad the stimulus incase windows go out of range.
    if np.any(win_start < 0) or \
            np.any(win_start >= (stimulus.shape[0] - total_len)):
        stimulus = np.pad(stimulus, ((total_len, total_len), (0, 0)),
                          'constant')
        # 3. Offset the windows, which is needed due to the padding.
        win_start += total_len
    # 4. Extract the slice.
    # The padded_stimulus is indexed by a list arrays of the form:
    #    (win_start[0], win_start[0]+1, win_start[0]+2, ..., win_start[0]+total_len)
    #    (win_start[1], win_start[1]+1, win_start[1]+2, ..., win_start[1]+total_len)
    #    ...
    #    (win_start[num_spikes-1], win_start[num_spikes-1]+1, win_start[num_spikes-1]+2, ..., win_start[num_spikes-1]+total_len)
    snippets = stimulus[np.asarray(win_start)[:, None] + np.arange(total_len)]
    return snippets


def labeled_spike_snippets(
        stimulus: np.ndarray, 
        response: pd.DataFrame,
        rec_name: str, 
        stimulus_sample_rate,
        sensor_sample_rate,
        snippet_len: int = 1000,
        snippet_pad: int = 200) -> Tuple[np.ndarray, np.ndarray]:
    """
    Caluclate spike snippets for a recording, paired with the cluster ids.

    Args:
        stimulus: the stimulus, sampled at `stimulus_sample_rate`.
        response: the response dataframe.
        rec_name: the name of the recording to extract snippets for.
        snippet_len: the length of the snippet.
        snippet_pad: the number of timesteps to include after the spike.
        stimulus_sample_rate: the sampling frequency of the stimulus.
        sensor_sample_rate: the sampling frequency of the sensor.
    
    Returns: two np.ndarrays as a tuple. The first element contains 
    the spike snippets, and the second element contains ids of the clusters.
    """
    response = response.xs(rec_name, level='Recording').reset_index(
        level='Cell index').drop('Kernel', axis=1, errors='ignore')

    def create_snippets(spikes) -> List[np.ndarray]:
        """Create the snippets for a list of spikes."""
        snips = []
        w = spike_snippets(stimulus, spikes.compressed(), stimulus_sample_rate,
                           sensor_sample_rate, snippet_len, snippet_pad)
        snips.extend(w)
        return snips

    # Gather the snippets and align them with a matching array of cluster ids.
    snippets = []
    cluster_ids = []
    for row_as_tuple in response.itertuples(index=False):
        cluster_id, spikes = row_as_tuple
        snips = create_snippets(spikes)
        cluster_ids.extend([cluster_id, ]*len(snips))
        snippets.extend(snips)
    snippets = np.stack(snippets)
    cluster_ids = np.array(cluster_ids)
    return snippets, cluster_ids


def generate_fake_spikes(spikes, num_fake_per_real,
                         target_sample_rate,
                         sensor_sample_rate,
                         min_dist_to_real_spike,
                         rng_seed=123) -> Tuple[np.ndarray, np.ndarray]:
    """
    Genrate fake spikes.

    Spread the spikes over the duration of the recording. Don't generate
    a fake spike too close to a real spike.

    Between every two spikes, `num_fake_per_real` spikes will be generated.
    Thus, total number of fake spikes will be 

        len(spikes) * `num_fake_per_real` -1.
    
    Some details:

        spikes:     |--------*------------*----------------------*-----------|
        intervals shown by |-----| in which we can create fake spikes:
                     |       * |--------| * |------------------| *           |

    There is a buffer between the spikes and the intervals, this is
    the `min_dist_to_real_spike`.

    Within the intervals, we will try to create `snippets_per_spike` number
    of fake spikes. From these spikes, snippets will be created. 

    The spikes are not chosen evenly within the intervals, but rather
    there is some jitter. This is what the `rng_seed` parameter is for.

    Args:
        spikes: a sorted (ascending) list of spikes, in sensor samples.
        num_fake_per_real: the number of fake spikes to generate between every
            two real spikes.
        target_sample_rate: the target sampling rate.
        sensor_sapmle_rate: the sampling rate of the sensor.
    """
    if len(spikes) < 2:
        raise ValueError(f'Need at least two spikes. Got ({len(spikes)}).')
    # Convert spikes to the target sample space.

    spikes = np.ndarray(spikes) * (target_sample_rate / sensor_sample_rate)

    intervals = np.stack(spikes[:-1] + min_dist_to_real_spike,
                         spikes[1:] - min_dist_to_real_spike)
    sub_intervals = []

    def split_interval(interval, n):
        """Split an interval into n subintervals, of roughly equal size."""
        d, m = divmod(interval, n)
        return [(i*d+min(i, m), (i+1)*d+min(i+1, m)) for i in range(n)]
    for i in intervals:
        sub_intervals.extend(split_interval(i, num_fake_per_real))

    rng = np.random.default_rng(rng_seed)
    fake_spikes = rng.random.integers(*np.stack(sub_intervals, axis=1),
                                      endpoints=True)
    return fake_spikes


def _save_recording_names(rec_list, save_dir) -> pathlib.Path:
    """
    Saves a list of recording names. 

    The list functions as an id to name map.
    """
    save_path = pathlib.Path(save_dir) / REC_NAMES_FILENAME
    # Create out file
    with save_path.open('wb') as f:
        pickle.dump(rec_list, f)
    return save_path


def _save_cluster_ids(cell_cluster_ids, rec_id, data_root_dir) -> pathlib.Path:
    """
    Saves a list of cluster ids. 

    The list functions as an id to name map.
    """
    parent_dir = pathlib.Path(data_root_dir) / str(rec_id)
    # Create if not exists
    parent_dir.mkdir(parents=False, exist_ok=True)
    save_path = pathlib.Path(data_root_dir) / \
        str(rec_id) / 'cluster_ids.pickle'
    # Create out file
    with save_path.open('wb') as f:
        pickle.dump(cell_cluster_ids, f)
    return save_path


def create_snippet_training_data(
        stimulus_pattern: np.ndarray,
        recorded_stimulus: pd.DataFrame,
        response: pd.DataFrame,
        save_dir,
        downsample_factor: int = 18,
        snippet_len: int = 1000,
        snippet_pad: int = 200,
        empty_snippets: float = 0,
        snippets_per_file: int = 1024):
    # TODO: WIP
    save_dir = pathlib.Path(save_dir)
    _save_recording_names(recording_names(response), save_dir)
    # TODO: only use stimulus 1? What is stimulus 7?
    by_rec = response.xs(1, level='Stimulus ID',
                         drop_level=True).groupby('Recording')
    for rec_id, (rec_name, df) in enumerate(by_rec):
        stimulus, freq = decompress_and_decimate(stimulus_pattern,
                recorded_stimulus, rec_name, downsample_factor)
        _write_rec_snippets(stimulus, df, rec_name, rec_id, save_dir,
                            snippet_len, snippet_pad, freq, empty_snippets,
                            snippets_per_file)


def _write_rec_snippets(
        stimulus: np.ndarray,
        response: pd.DataFrame,
        rec_name: str,
        rec_id: int,
        data_root_dir,
        snippet_len: int,
        snippet_pad: int,
        stimulus_freq: float,
        empty_snippets: float,
        snippets_per_file: int):
    """
    Generate, shuffle and write the snippets for a single recording.

    The snippets may be split across multiple files.
    """
    # Create directory for recording.
    rec_dir = pathlib.Path(data_root_dir) / str(rec_id)
    rec_dir.mkdir(parents=False, exist_ok=True)
    # Extract the stimulus snippet around the spikes.
    snippets, cluster_ids = labeled_spike_snippets(
        stimulus, response, rec_name, stimulus_freq, ELECTRODE_FREQ, 
        snippet_len, snippet_pad)
    # Save the cluster ids.
    _save_cluster_ids(cluster_ids, rec_id, data_root_dir)
    assert len(cluster_ids) == len(snippets)
    # Shuffle the cluster ids and snippets together.
    rng = np.random.default_rng(RNG_SEED)
    shuffle_idxs = np.arange(len(cluster_ids))
    rng.shuffle(shuffle_idxs)
    cluster_ids = cluster_ids[shuffle_idxs]
    snippets = snippets[shuffle_idxs]
    # Fill array with rec_id
    rec_ids = np.full(len(cluster_ids), rec_id)
    # Create 2 column array, recording ids and cluster ids.
    # ([1, 1, 1, 1], [1, 2, 3, 4]) -> [[1, 1], [1, 2], [1, 3], [1, 4]]
    rec_clusters = np.vstack((rec_ids, cluster_ids)).T

    # Save the snippets and ids across multiple files.
    def save_splits():
        ids_split = np.array_split(rec_clusters, snippets_per_file)
        snippets_split = np.array_split(snippets, snippets_per_file)
        assert len(ids_split) == len(snippets_split)
        for i, (ids, ks) in enumerate(zip(ids_split, snippets_split)):
            _write_snippet_data_part(ids, ks, rec_dir, i)
    save_splits()


def _write_snippet_data_part(ids, snippets, save_dir, part_id):
    """
    Saves a subset of the spike snippets.

    This is used internally to spread the snippets data over multiple files.
    """
    ids_filename = IDS_FILE_PATTERN.format(part=part_id)
    win_filename = SNIPPET_FILE_PATTERN.format(part=part_id)
    ids_save_path = pathlib.Path(save_dir) / ids_filename
    win_save_path = pathlib.Path(save_dir) / win_filename
    with ids_save_path.open('wb') as f:
        np.save(f, ids)
    with win_save_path.open('wb') as f:
        np.save(f, snippets)
    return ids_save_path, win_save_path



import torch
import torch.nn as nn
import torch.nn.functional as F


class GanglionAsCNN(nn.Module):
    LED_CHANNELS = 4
    NUM_CLUSTERS = 1

    def __init__(self, input_len, input_freq, kernel_len, output_offset):
        """
        Args:
            input_len: length of input.
            input_freq: sampling frequency of input.
            kernel_len: length of kernel to use.
            output_offset: number of time steps after the kernel to predict.
                This is measured from the last element of the kernel, so 
                an offset of zero means that the output predicts the response
                present in the same timestep as the last elment of the kernel.
                This property is similar to the kernel pad length used when
                collecting spike snippets.
        """
        super(GanglionAsCNN, self).__init__()
        self.input_len = input_len
        self.input_freq = input_freq
        self.output_offset = output_offset
        # Input is the LED stimulus and the cell cluster's response.
        self.num_input_channels = self.LED_CHANNELS + self.NUM_CLUSTERS
        
        self.cnn1 = nn.Conv1d(
                in_channels=self.num_input_channels,
                out_channels=self.NUM_CLUSTERS,
                kernel_size=kernel_len,
                stride=1, padding=0, dilation=1)

    def forward(self, stimulus, response):
        x = torch.stack([stimulus, response], dim=1)
        x = self.cnn1(x)
        x = F.log_softmax(x, dim=1)
        return x

    def loss(self, pred, response):
        y = response[self.kernel_len + self.output_offset:]
        # TODO: how to handle the issue that this loss will prioritize 
        # no-spike responses, given that they are so numerous? Maybe we can 
        # use a weight matrix that weights the no-spike loss by the 1/ratio of
        # no-spike:spike.
        loss = F.binary_cross_entropy(pred, y)
        return loss


def train_model(model, data, epochs):
    for epoch in range(epochs):
        for i, (stimulus, response) in enumerate(data):
            pred = model(stimulus, response)
            loss = model.loss(pred, response)
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
            if i % 100 == 0:
                print(f'Epoch {epoch}, iteration {i}, loss {loss}')
    return model


if __name__ == '__main__':
    model = GanglionAsCNN(input_len=100, input_freq=1000,
                          kernel_len=10, output_offset=0)

    train_madel(model, data, epochs=10)
